# -*- coding: UTF-8 -*-
'''
Simple modułe for fetching data from Polish REGON database.
As all Python SOAP libraries I know have problem accessing this service,
I reimplemented querying with plain requests.

Created on 17 lip 2015

@author: Michał Węgrzynek
'''
from __future__ import unicode_literals, print_function
import logging
import email
import sys
import warnings


import requests
from lxml import etree, objectify


from .envelopes import (
    LOGIN_ENVELOPE,
    LOGOUT_ENVELOPE,
    SEARCH_ENVELOPE,
    FULL_REPORT_ENVELOPE,
    GET_CAPTCHA_ENVELOPE,
    CHECK_CAPTCHA_ENVELOPE
)


py3 = sys.version_info >= (3,)
log = logging.getLogger(__name__)


namespaces = {
    'bir': 'http://CIS/BIR/PUBL/2014/07',
    'pb': 'http://CIS/BIR/2014/07'
}


detailed_report_names_map = {
    '6': 'PublDaneRaportPrawna',
    '1': 'PublDaneRaportDzialalnoscFizycznejCeidg',
    '2': 'PublDaneRaportDzialalnoscFizycznejRolnicza',
    '3': 'PublDaneRaportDzialalnoscFizycznejPozostala',
    '4': 'PublDaneRaportDzialalnoscFizycznejWKrupgn'
}

extended_report_names_map = {
    'LP': 'PublDaneRaportLokalnaPrawnej',
    'LF': 'PublDaneRaportLokalnaFizycznej',
    'F': 'PublDaneRaportFizycznaOsoba'
}


def get_message_element(message, payload_num, path):
    resp = message.get_payload(payload_num).get_payload()
    return etree.fromstring(resp).xpath(path, namespaces=namespaces)


class REGONAPIError(RuntimeError):
    pass


class REGONAPI(object):

    def __init__(self, service_url):
        self.service_url = service_url
        self.sid = None

    def call(self, envelope, **args):
        '''
        Calls an API's method (descibed by the envelope)
        '''
        data = envelope.format(api=self, **args)
        log.debug('Data to be posted: %s', data)
        res = requests.post(
            self.service_url,
            stream=True,
            data=data,
            headers={
                'Content-Type': 'application/soap+xml; charset=utf-8',
                'sid': self.sid if self.sid else None
            }
        )
        mimemsg = '\r\n'.join(
            '{0}: {1}'.format(key, val) for key, val in res.headers.items()
        )
        mimemsg += '\r\n' + res.text

        if not py3:
            mimemsg = mimemsg.encode('UTF-8')

        mesg = email.message_from_string(mimemsg)
        assert mesg.is_multipart, 'Response is not multipart.'

        return mesg

    def login(self, user_key):
        '''
        Logs in to the REGON API Service
        (uses API key provided by the REGON administrators)
        '''
        mesg = self.call(LOGIN_ENVELOPE, user_key=user_key)

        result = get_message_element(mesg, 0, '//bir:ZalogujResult/text()')
        if not result:
            self.sid = None
            raise REGONAPIError('Login failed.')

        self.sid = result[0]

        return self.sid

    def logout(self):
        '''
        Ends API session
        '''
        if not self.sid:
            raise REGONAPIError('Not logged in.')

        mesg = self.call(LOGOUT_ENVELOPE)
        result = get_message_element(mesg, 0, '//bir:WylogujResult/text()')

        if not result or result[0] != 'true':
            raise REGONAPIError('Logout failed.')

        return True

    def get_captcha(self):
        '''
        Gets CAPTCHA for human verification;
        Returns decoded image data or false, if showing CAPTCHA is not needed.
        Deprecated.
        '''
        warnings.warn(
            'No longer needed.',
            DeprecationWarning
        )

        mesg = self.call(GET_CAPTCHA_ENVELOPE)
        result = get_message_element(
            mesg,
            0,
            '//pb:PobierzCaptchaResult/text()'
        )
        if not result:
            raise REGONAPIError('Getting CAPTCHA failed.')

        resp = result[0]
        return resp if resp else False

    def check_captcha(self, captcha):
        '''
        Sends the human recognized CAPTCHA string for verification
        '''
        warnings.warn(
            'No longer needed.',
            DeprecationWarning
        )

        mesg = self.call(CHECK_CAPTCHA_ENVELOPE, captcha=captcha)
        result = get_message_element(
            mesg,
            0,
            '//pb:SprawdzCaptchaResult/text()'
        )
        if not result:
            raise REGONAPIError('Checking CAPTCHA failed.')

        resp = result[0]
        return resp == 'true'

    def search(
        self,
        nip=None,
        regon=None,
        krs=None,
        nips=None,
        regons=None,
        krss=None,
        detailed=False
    ):
        if not (regon or nip or krs or regons or nips or krss):
            raise REGONAPIError(
                'You have to pass at least one of: '
                'nip(s), regon(s) or krs(s) parameters.'
            )

        param = ''
        if nip:
            param += '<dat:Nip>{0}</dat:Nip>'.format(nip)

        if regon:
            param += '<dat:Regon>{0}</dat:Regon>'.format(regon)

        if krs:
            param += '<dat:Krs>{0}</dat:Krs>'.format(krs)

        if nips:
            nip_str = ''.join(nips)
            assert len(nip_str) % 10 == 0, \
                'All NIPs should be 10 character strings.'

            param += '<dat:Nipy>{0}</dat:Nipy>'.format(nip_str)

        if krss:
            krs_str = ''.join(krss)
            assert len(krs_str) % 10 == 0, \
                'All KRSs should be 10 character strings.'
            param += '<dat:Krsy>{0}</dat:Krsy>'.format(krs_str)

        if regons:
            regons_str = ''.join(regons)
            rl = len(regons_str)

            if rl % 9 == 0:
                param += '<dat:Regony9zn>{0}</dat:Regony9zn>'.format(
                    regons_str
                )
            elif rl % 14 == 0:
                param += '<dat:Regony14zn>{0}</dat:Regony14zn>'.format(
                    regons_str
                )
            else:
                raise AssertionError(
                    'All REGONs should be either 9 or 14 character strings.'
                )

        mesg = self.call(SEARCH_ENVELOPE, param=param)
        result = get_message_element(mesg, 0, '//bir:DaneSzukajResult/text()')
        if not result:
            raise REGONAPIError('Search failed.')

        search_results = list(objectify.fromstring(result[0]).dane)

        if not detailed:
            return search_results
        else:
            detailed_data = []
            for rs in search_results:

                # Sometimes the leading zeros get lost
                correct_regon = '{0:014}'.format(int(rs.Regon)) \
                    if len(str(rs.Regon)) not in (9, 14) else rs.Regon

                correct_report_name = detailed_report_names_map.get(
                    str(rs.SilosID)
                )

                rs.detailed = self.full_report(
                    correct_regon,
                    correct_report_name
                )

                # Some types of organizations have additional detailed
                # reports
                extended_report_name = extended_report_names_map.get(rs.Typ)
                if extended_report_name:
                    add_data = self.full_report(
                        correct_regon,
                        extended_report_name
                    )
                    rs.detailed.extend(add_data.getchildren())

                detailed_data.append(rs)

            return detailed_data

    def full_report(self, regon, report_name):
        result = None

        mesg = get_message_element(
            self.call(
                FULL_REPORT_ENVELOPE,
                regon=regon,
                report_name=report_name
            ),
            0,
            '//bir:DanePobierzPelnyRaportResult/text()'
        )

        if mesg:
            result = objectify.fromstring(
                mesg[0]
            )
            result = result[0].dane
        else:
            result = objectify.Element("detailed")

        return result



# -*- encoding: UTF-8 -*-
import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()

requires = [
    'setuptools',
    'requests',
    'lxml'
]

version = '1.0.4'

setup(
    name='litex.regon',
    version=version,
    description='An API for accessing a Polish REGON database',
    long_description=README,
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Database :: Front-Ends"
    ],
    author='Michal Wegrzynek',
    author_email='mwegrzynek@litex.pl',
    url='http://litexservice.pl',
    license='GPL',
    keywords='regon soap database',
    package_dir={'': '.'},
    namespace_packages=['litex'],
    packages=find_packages('.'),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
)



# -*- coding: UTF-8 -*-
'''
SOAP Envelopes for REGON API
Created on 17 lip 2015

@author: Michał Węgrzynek
'''

LOGIN_ENVELOPE = '''\
<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:ns="http://CIS/BIR/PUBL/2014/07">
<soap:Header xmlns:wsa="http://www.w3.org/2005/08/addressing">
<wsa:To>{api.service_url}</wsa:To>
<wsa:Action>http://CIS/BIR/PUBL/2014/07/IUslugaBIRzewnPubl/Zaloguj</wsa:Action>
</soap:Header>
   <soap:Body>
      <ns:Zaloguj>
         <ns:pKluczUzytkownika>{user_key}</ns:pKluczUzytkownika>
      </ns:Zaloguj>
   </soap:Body>
</soap:Envelope>
'''

LOGOUT_ENVELOPE = '''\
<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:ns="http://CIS/BIR/PUBL/2014/07">
<soap:Header xmlns:wsa="http://www.w3.org/2005/08/addressing">
<wsa:To>{api.service_url}</wsa:To>
<wsa:Action>http://CIS/BIR/PUBL/2014/07/IUslugaBIRzewnPubl/Wyloguj</wsa:Action>
</soap:Header>
   <soap:Body>
      <ns:Wyloguj>
         <ns:pIdentyfikatorSesji>{api.sid}</ns:pIdentyfikatorSesji>
      </ns:Wyloguj>
   </soap:Body>
</soap:Envelope>
'''

GET_CAPTCHA_ENVELOPE = '''\
<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:ns="http://CIS/BIR/2014/07">
<soap:Header xmlns:wsa="http://www.w3.org/2005/08/addressing">
<wsa:To>{api.service_url}</wsa:To>
<wsa:Action>http://CIS/BIR/2014/07/IUslugaBIR/PobierzCaptcha</wsa:Action>
</soap:Header>
   <soap:Body>
      <ns:PobierzCaptcha/>
   </soap:Body>
</soap:Envelope>
'''

CHECK_CAPTCHA_ENVELOPE = '''\
<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:ns="http://CIS/BIR/2014/07">
<soap:Header xmlns:wsa="http://www.w3.org/2005/08/addressing">
<wsa:To>{api.service_url}</wsa:To>
<wsa:Action>http://CIS/BIR/2014/07/IUslugaBIR/SprawdzCaptcha</wsa:Action>
</soap:Header>
   <soap:Body>
      <ns:SprawdzCaptcha>
         <ns:pCaptcha>{captcha}</ns:pCaptcha>
      </ns:SprawdzCaptcha>
   </soap:Body>
</soap:Envelope>
'''

SEARCH_ENVELOPE = '''\
<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:ns="http://CIS/BIR/PUBL/2014/07" xmlns:dat="http://CIS/BIR/PUBL/2014/07/DataContract">
<soap:Header xmlns:wsa="http://www.w3.org/2005/08/addressing">
<wsa:To>{api.service_url}</wsa:To>
<wsa:Action>http://CIS/BIR/PUBL/2014/07/IUslugaBIRzewnPubl/DaneSzukaj</wsa:Action>
</soap:Header>
   <soap:Body>
      <ns:DaneSzukaj>
         <ns:pParametryWyszukiwania>
            {param}
         </ns:pParametryWyszukiwania>
      </ns:DaneSzukaj>
   </soap:Body>
</soap:Envelope>
'''

FULL_REPORT_ENVELOPE = '''\
<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:ns="http://CIS/BIR/PUBL/2014/07">
<soap:Header xmlns:wsa="http://www.w3.org/2005/08/addressing">
<wsa:To>{api.service_url}</wsa:To>
<wsa:Action>http://CIS/BIR/PUBL/2014/07/IUslugaBIRzewnPubl/DanePobierzPelnyRaport</wsa:Action>
</soap:Header>
  <soap:Body>
      <ns:DanePobierzPelnyRaport>
         <ns:pRegon>{regon}</ns:pRegon>
         <ns:pNazwaRaportu>{report_name}</ns:pNazwaRaportu>
      </ns:DanePobierzPelnyRaport>
   </soap:Body>
</soap:Envelope>
'''



# -*- coding: UTF-8 -*-
'''
Basic tests of REGON API

Created on 17 lip 2015

@author: Michał Wegrzynek <mwegrzynek@litex.pl>
'''
from __future__ import unicode_literals
import os
import sys


import pytest
from lxml import etree


from litex.regon import REGONAPIError, REGONAPI


py3 = sys.version_info >= (3,)


USER_KEY = os.environ.get('LR_USER_KEY')
SERVICE_URL = os.environ.get('LR_SERVICE_URL')
TEST_NIP_SP = os.environ.get('LR_TEST_NIP_SP')
TEST_REGON_SP = int(os.environ.get('LR_TEST_REGON_SP'))
TEST_NAME_SP = os.environ.get('LR_TEST_NAME_SP')
TEST_NIP_CP = os.environ.get('LR_TEST_NIP_CP')
TEST_REGON_CP = int(os.environ.get('LR_TEST_REGON_CP'))
TEST_NAME_CP = os.environ.get('LR_TEST_NAME_CP')
if not py3:
    TEST_NAME_SP = unicode(TEST_NAME_SP, 'UTF-8')
    TEST_NAME_CP = unicode(TEST_NAME_CP, 'UTF-8')


@pytest.fixture
def api():
    return REGONAPI(SERVICE_URL)


@pytest.fixture
def li_api(api):
    '''Logged in API'''
    api.login(USER_KEY)
    return api


def test_login_correct_user_key(api):
    sid = api.login(USER_KEY)
    assert len(sid) == 20


def test_login_incorrect_user_key(api):
    with pytest.raises(REGONAPIError):
        api.login('dummykey')


def test_logout_without_login(api):
    with pytest.raises(REGONAPIError):
        api.logout()


def test_logout(api):
    api.login(USER_KEY)
    api.logout()


def test_search_no_params(li_api):
    with pytest.raises(REGONAPIError):
        li_api.search()


def test_search_sole_proprietorship(li_api):
    result = li_api.search(nip=TEST_NIP_SP)
    assert result[0].Regon == TEST_REGON_SP
    assert result[0].Nazwa == TEST_NAME_SP
    li_api.logout()


def test_search_sole_proprietorship_detailed(li_api):
    result = li_api.search(nip=TEST_NIP_SP, detailed=True)[0]
    assert result[0].Regon == TEST_REGON_SP
    assert str(result[0].detailed.fiz_nip) == TEST_NIP_SP
    assert getattr(
        result.detailed,
        'fiz_adSiedzNumerNieruchomosci'
    ) is not None
    li_api.logout()


def test_search_corporation(li_api):
    result = li_api.search(nip=TEST_NIP_CP)
    assert result[0].Regon == TEST_REGON_CP
    assert result[0].Nazwa == TEST_NAME_CP
    li_api.logout()


def test_search_corporation_detailed(li_api):
    result = li_api.search(nip=TEST_NIP_CP, detailed=True)[0]
    assert result.Regon == TEST_REGON_CP
    assert getattr(
        result.detailed,
        'praw_adSiedzNumerNieruchomosci'
    ) is not None
    li_api.logout()


def test_search_multiple_nips(li_api):
    result = li_api.search(nips=[TEST_NIP_CP, TEST_NIP_SP])
    assert len(result) >= 2
    assert result[-1].Regon == TEST_REGON_CP
    assert result[0].Regon == TEST_REGON_SP
    li_api.logout()


def test_issue_3_charities(li_api):
    result = li_api.search(
        regons=[
            '01210292600037',
            '01210292600051',
            '01210292600076'
        ],
        detailed=True
    )

    assert len(result) == 3

    for res in result:
        assert res.detailed.lokpraw_regon14 == res.Regon

    li_api.logout()